@@ -31,6 +31,11 @@
 /*	VSTRING *why;
 /*	int	lflags;
 /*	unsigned *ltype;
+/*
+/*	int	dns_get_h_errno()
+/*
+/*	void	dns_set_h_errno(
+/*	int	errval)
 /* AUXILIARY FUNCTIONS
 /*	extern int var_dns_ncache_ttl_fix;
 /*
@@ -83,6 +88,10 @@
 /*	an invalid name is reported as a DNS_INVAL result, while
 /*	malformed replies are reported as transient errors.
 /*
+/*	dns_get_h_errno() returns the last error, and dns_set_h_errno()
+/*	sets it. This deprecates usage of the global h_errno variable.
+/*	We should not rely on that being updated.
+/*
 /*	dns_lookup_l() and dns_lookup_v() allow the user to specify
 /*	a list of resource types.
 /*
@@ -171,6 +180,12 @@
 /*	Pointer to storage for the reply RCODE value. This gives
 /*	more detailed information than DNS_FAIL, DNS_RETRY, etc.
 /* DIAGNOSTICS
+/*	If DNSSEC validation is requested but the response is not
+/*	DNSSEC validated, dns_lookup() will send a one-time probe
+/*	query as configured with the \fBdnssec_probe\fR configuration
+/*	parameter, and will log a warning when the probe response
+/*	was not DNSSEC validated.
+/* .PP
 /*	dns_lookup() returns one of the following codes and sets the
 /*	\fIwhy\fR argument accordingly:
 /* .IP DNS_OK
@@ -283,13 +298,17 @@ typedef struct DNS_REPLY {
 #define INET6_ADDR_LEN	16		/* XXX */
 
  /*
-  * Use the theadsafe resolver API if available, not because it is theadsafe,
-  * but because it has more functionality.
+  * Use the threadsafe resolver API if available, not because it is
+  * theadsafe, but because it has more functionality.
   */
-#ifdef USE_NRES_INIT
+#ifdef USE_RES_NCALLS
 static struct __res_state dns_res_state;
 
-#define DNS_RES_H_ERRNO(statp)	((statp)->res_h_errno)
+#define DNS_RES_NINIT		res_ninit
+#define DNS_RES_NMKQUERY	res_nmkquery
+#define DNS_RES_NSEARCH		res_nsearch
+#define DNS_RES_NSEND		res_nsend
+#define DNS_GET_H_ERRNO(statp)	((statp)->res_h_errno)
 
  /*
   * Alias new resolver API calls to the legacy resolver API which stores
@@ -297,30 +316,26 @@ static struct __res_state dns_res_state;
   */
 #else
 #define dns_res_state		_res
-#undef	res_ninit
-#undef	res_nmkquery
-#undef	res_nsearch
-#undef	res_nsend
-#define res_ninit(statp)	res_init()
-#define res_nmkquery(statp, op, dname, class, type, data, datalen, \
+#define DNS_RES_NINIT(statp)	res_init()
+#define DNS_RES_NMKQUERY(statp, op, dname, class, type, data, datalen, \
 		newrr, buf, buflen) \
 	res_mkquery((op), (dname), (class), (type), (data), (datalen), \
 		(newrr), (buf), (buflen))
-#define res_nsearch(statp, dname, class, type, answer, anslen) \
+#define DNS_RES_NSEARCH(statp, dname, class, type, answer, anslen) \
 	res_search((dname), (class), (type), (answer), (anslen))
-#define res_nsend(statp, msg, msglen, answer, anslen) \
+#define DNS_RES_NSEND(statp, msg, msglen, answer, anslen) \
 	res_send((msg), (msglen), (answer), (anslen))
-#define DNS_RES_H_ERRNO(statp)	(h_errno)
+#define DNS_GET_H_ERRNO(statp)	(h_errno)
 #endif
 
 #ifdef USE_SET_H_ERRNO
-#define DNS_RES_SET_H_ERRNO(statp, err)	(set_h_errno(err))
+#define DNS_SET_H_ERRNO(statp, err)	(set_h_errno(err))
 #else
-#define DNS_RES_SET_H_ERRNO(statp, err)	(DNS_RES_H_ERRNO(statp) = (err))
+#define DNS_SET_H_ERRNO(statp, err)	(DNS_GET_H_ERRNO(statp) = (err))
 #endif
 
  /*
-  * To improve postscreen's whitelisting support, we need to know how long a
+  * To improve postscreen's allowlisting support, we need to know how long a
   * DNSBL "not found" answer is valid. The 2010 implementation assumed it was
   * valid for 3600 seconds. That is too long by 2015 standards.
   * 
@@ -349,9 +364,9 @@ static struct __res_state dns_res_state;
   */
 #ifdef HAVE_RES_SEND
 
-/* dns_res_query - a res_query() clone that can return negative replies */
+/* dns_neg_query - a res_query() clone that can return negative replies */
 
-static int dns_res_query(const char *name, int class, int type,
+static int dns_neg_query(const char *name, int class, int type,
 			         unsigned char *answer, int anslen)
 {
     unsigned char msg_buf[MAX_DNS_QUERY_SIZE];
@@ -380,36 +395,36 @@ static int dns_res_query(const char *name, int class, int type,
 #define NO_MKQUERY_DATA_LEN     ((int) 0)
 #define NO_MKQUERY_NEWRR        ((unsigned char *) 0)
 
-    if ((len = res_nmkquery(&dns_res_state,
-			    QUERY, name, class, type, NO_MKQUERY_DATA_BUF,
-			    NO_MKQUERY_DATA_LEN, NO_MKQUERY_NEWRR,
-			    msg_buf, sizeof(msg_buf))) < 0) {
-	DNS_RES_SET_H_ERRNO(&dns_res_state, NO_RECOVERY);
+    if ((len = DNS_RES_NMKQUERY(&dns_res_state,
+			      QUERY, name, class, type, NO_MKQUERY_DATA_BUF,
+				NO_MKQUERY_DATA_LEN, NO_MKQUERY_NEWRR,
+				msg_buf, sizeof(msg_buf))) < 0) {
+	DNS_SET_H_ERRNO(&dns_res_state, NO_RECOVERY);
 	if (msg_verbose)
 	    msg_info("res_nmkquery() failed");
 	return (len);
-    } else if ((len = res_nsend(&dns_res_state,
-				msg_buf, len, answer, anslen)) < 0) {
-	DNS_RES_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
+    } else if ((len = DNS_RES_NSEND(&dns_res_state,
+				    msg_buf, len, answer, anslen)) < 0) {
+	DNS_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
 	if (msg_verbose)
 	    msg_info("res_nsend() failed");
 	return (len);
     } else {
 	switch (reply_header->rcode) {
 	case NXDOMAIN:
-	    DNS_RES_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
+	    DNS_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
 	    break;
 	case NOERROR:
 	    if (reply_header->ancount != 0)
-		DNS_RES_SET_H_ERRNO(&dns_res_state, 0);
+		DNS_SET_H_ERRNO(&dns_res_state, 0);
 	    else
-		DNS_RES_SET_H_ERRNO(&dns_res_state, NO_DATA);
+		DNS_SET_H_ERRNO(&dns_res_state, NO_DATA);
 	    break;
 	case SERVFAIL:
-	    DNS_RES_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
+	    DNS_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
 	    break;
 	default:
-	    DNS_RES_SET_H_ERRNO(&dns_res_state, NO_RECOVERY);
+	    DNS_SET_H_ERRNO(&dns_res_state, NO_RECOVERY);
 	    break;
 	}
 	return (len);
@@ -418,9 +433,9 @@ static int dns_res_query(const char *name, int class, int type,
 
 #endif
 
-/* dns_res_search - res_search() that can return negative replies */
+/* dns_neg_search - res_search() that can return negative replies */
 
-static int dns_res_search(const char *name, int class, int type,
+static int dns_neg_search(const char *name, int class, int type,
 	               unsigned char *answer, int anslen, int keep_notfound)
 {
     int     len;
@@ -443,19 +458,19 @@ static int dns_res_search(const char *name, int class, int type,
     if (keep_notfound)
 	/* Prepare for returning a null-padded server reply. */
 	memset(answer, 0, anslen);
-    len = res_nsearch(&dns_res_state, name, class, type, answer, anslen);
+    len = DNS_RES_NSEARCH(&dns_res_state, name, class, type, answer, anslen);
     /* Begin API creep workaround. */
-    if (len < 0 && DNS_RES_H_ERRNO(&dns_res_state) == 0) {
-	DNS_RES_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
+    if (len < 0 && DNS_GET_H_ERRNO(&dns_res_state) == 0) {
+	DNS_SET_H_ERRNO(&dns_res_state, TRY_AGAIN);
 	msg_warn("res_nsearch(state, \"%s\", %d, %d, %p, %d) returns %d"
 		 " with h_errno==0 -- setting h_errno=TRY_AGAIN",
 		 name, class, type, answer, anslen, len);
     }
     /* End API creep workaround. */
     if (len > 0) {
-	DNS_RES_SET_H_ERRNO(&dns_res_state, 0);
+	DNS_SET_H_ERRNO(&dns_res_state, 0);
     } else if (keep_notfound
-	       && NOT_FOUND_H_ERRNO(DNS_RES_H_ERRNO(&dns_res_state))) {
+	       && NOT_FOUND_H_ERRNO(DNS_GET_H_ERRNO(&dns_res_state))) {
 	/* Expect to return a null-padded server reply. */
 	len = anslen;
     }
@@ -484,7 +499,7 @@ static int dns_query(const char *name, int type, unsigned flags,
      * Initialize the name service.
      */
     if ((dns_res_state.options & RES_INIT) == 0
-	&& res_ninit(&dns_res_state) < 0) {
+	&& DNS_RES_NINIT(&dns_res_state) < 0) {
 	if (why)
 	    vstring_strcpy(why, "Name service initialization failure");
 	return (DNS_FAIL);
@@ -504,7 +519,7 @@ static int dns_query(const char *name, int type, unsigned flags,
      */
 #define XTRA_FLAGS (RES_USE_EDNS0 | RES_TRUSTAD)
 
-    if (flags & RES_USE_DNSSEC)
+    if (DNS_WANT_DNSSEC_VALIDATION(flags))
 	flags |= (RES_USE_EDNS0 | RES_TRUSTAD);
 
     /*
@@ -534,34 +549,36 @@ static int dns_query(const char *name, int type, unsigned flags,
 	dns_res_state.options |= flags;
 	if (keep_notfound && var_dns_ncache_ttl_fix) {
 #ifdef HAVE_RES_SEND
-	    len = dns_res_query((char *) name, C_IN, type, reply->buf,
+	    len = dns_neg_query((char *) name, C_IN, type, reply->buf,
 				reply->buf_len);
 #else
 	    var_dns_ncache_ttl_fix = 0;
 	    msg_warn("system library does not support %s=yes"
 		     " -- ignoring this setting", VAR_DNS_NCACHE_TTL_FIX);
-	    len = dns_res_search((char *) name, C_IN, type, reply->buf,
+	    len = dns_neg_search((char *) name, C_IN, type, reply->buf,
 				 reply->buf_len, keep_notfound);
 #endif
 	} else {
-	    len = dns_res_search((char *) name, C_IN, type, reply->buf,
+	    len = dns_neg_search((char *) name, C_IN, type, reply->buf,
 				 reply->buf_len, keep_notfound);
 	}
 	dns_res_state.options &= ~flags;
 	dns_res_state.options |= saved_options;
 	reply_header = (HEADER *) reply->buf;
 	reply->rcode = reply_header->rcode;
-	if (DNS_RES_H_ERRNO(&dns_res_state) != 0) {
+	if ((reply->dnssec_ad = !!reply_header->ad) != 0)
+	    DNS_SEC_STATS_SET(DNS_SEC_FLAG_AVAILABLE);
+	if (DNS_GET_H_ERRNO(&dns_res_state) != 0) {
 	    if (why)
 		vstring_sprintf(why, "Host or domain name not found. "
 				"Name service error for name=%s type=%s: %s",
-				name, dns_strtype(type), 
-				dns_strerror(DNS_RES_H_ERRNO(&dns_res_state)));
+				name, dns_strtype(type),
+			     dns_strerror(DNS_GET_H_ERRNO(&dns_res_state)));
 	    if (msg_verbose)
 		msg_info("dns_query: %s (%s): %s",
-			 name, dns_strtype(type), 
-			dns_strerror(DNS_RES_H_ERRNO(&dns_res_state)));
-	    switch (DNS_RES_H_ERRNO(&dns_res_state)) {
+			 name, dns_strtype(type),
+			 dns_strerror(DNS_GET_H_ERRNO(&dns_res_state)));
+	    switch (DNS_GET_H_ERRNO(&dns_res_state)) {
 	    case NO_RECOVERY:
 		return (DNS_FAIL);
 	    case HOST_NOT_FOUND:
@@ -591,7 +608,7 @@ static int dns_query(const char *name, int type, unsigned flags,
      */
     if (len < 0)
 	msg_panic("dns_query: bad length %d (h_errno=%s)",
-		  len, dns_strerror(DNS_RES_H_ERRNO(&dns_res_state)));
+		  len, dns_strerror(DNS_GET_H_ERRNO(&dns_res_state)));
 
     /*
      * Paranoia.
@@ -604,13 +621,8 @@ static int dns_query(const char *name, int type, unsigned flags,
 
     /*
      * Initialize the reply structure. Some structure members are filled on
-     * the fly while the reply is being parsed.  Coerce AD bit to boolean.
+     * the fly while the reply is being parsed.
      */
-#if RES_USE_DNSSEC != 0
-    reply->dnssec_ad = (flags & RES_USE_DNSSEC) ? !!reply_header->ad : 0;
-#else
-    reply->dnssec_ad = 0;
-#endif
     SET_HAVE_DNS_REPLY_PACKET(reply, len);
     reply->query_start = reply->buf + sizeof(HEADER);
     reply->answer_start = 0;
@@ -625,13 +637,13 @@ static int dns_query(const char *name, int type, unsigned flags,
      * Future proofing. If this reaches the panic call, then some code change
      * introduced a bug.
      */
-    if (DNS_RES_H_ERRNO(&dns_res_state) == 0) {
+    if (DNS_GET_H_ERRNO(&dns_res_state) == 0) {
 	return (DNS_OK);
     } else if (keep_notfound) {
 	return (DNS_NOTFOUND);
     } else {
 	msg_panic("dns_query: unexpected reply status: %s",
-		  dns_strerror(DNS_RES_H_ERRNO(&dns_res_state)));
+		  dns_strerror(DNS_GET_H_ERRNO(&dns_res_state)));
     }
 }
 
@@ -928,7 +940,9 @@ static int dns_get_answer(const char *orig_name, DNS_REPLY *reply, int type,
 	    CORRUPT(DNS_RETRY);
 	if ((status = dns_get_fixed(pos, &fixed)) != DNS_OK)
 	    CORRUPT(status);
-	if (!valid_rr_name(rr_name, "resource name", fixed.type, reply))
+	if (strcmp(orig_name, ".") == 0 && *rr_name == 0)
+	     /* Allow empty response name for root queries. */ ;
+	else if (!valid_rr_name(rr_name, "resource name", fixed.type, reply))
 	    CORRUPT(DNS_INVAL);
 	if (fqdn)
 	    vstring_strcpy(fqdn, rr_name);
@@ -1009,21 +1023,21 @@ int     dns_lookup_x(const char *name, unsigned type, unsigned flags,
 			    name);
 	if (rcode)
 	    *rcode = NXDOMAIN;
-	DNS_RES_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
+	DNS_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
 	return (DNS_NOTFOUND);
     }
 
     /*
      * The Linux resolver misbehaves when given an invalid domain name.
      */
-    if (!valid_hostname(name, DONT_GRIPE)) {
+    if (strcmp(name, ".") && !valid_hostname(name, DONT_GRIPE)) {
 	if (why)
 	    vstring_sprintf(why,
 		   "Name service error for %s: invalid host or domain name",
 			    name);
 	if (rcode)
 	    *rcode = NXDOMAIN;
-	DNS_RES_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
+	DNS_SET_H_ERRNO(&dns_res_state, HOST_NOT_FOUND);
 	return (DNS_NOTFOUND);
     }
 
@@ -1053,6 +1067,10 @@ int     dns_lookup_x(const char *name, unsigned type, unsigned flags,
 		(void) dns_get_answer(orig_name, &reply, T_SOA, rrlist, fqdn,
 				      cname, c_len, &maybe_secure);
 	    }
+	    if (DNS_WANT_DNSSEC_VALIDATION(flags)
+		&& !DNS_SEC_STATS_TEST(DNS_SEC_FLAG_AVAILABLE | \
+				       DNS_SEC_FLAG_DONT_PROBE))
+		dns_sec_probe(flags);		/* XXX Clobbers 'reply' */
 	    return (status);
 	}
 
@@ -1062,6 +1080,10 @@ int     dns_lookup_x(const char *name, unsigned type, unsigned flags,
 	 */
 	status = dns_get_answer(orig_name, &reply, type, rrlist, fqdn,
 				cname, c_len, &maybe_secure);
+	if (DNS_WANT_DNSSEC_VALIDATION(flags)
+	    && !DNS_SEC_STATS_TEST(DNS_SEC_FLAG_AVAILABLE | \
+				   DNS_SEC_FLAG_DONT_PROBE))
+	    dns_sec_probe(flags);		/* XXX Clobbers 'reply' */
 	switch (status) {
 	default:
 	    if (why)
@@ -1073,7 +1095,7 @@ int     dns_lookup_x(const char *name, unsigned type, unsigned flags,
 	    if (why)
 		vstring_sprintf(why, "Domain %s does not accept mail (nullMX)",
 				name);
-	    DNS_RES_SET_H_ERRNO(&dns_res_state, NO_DATA);
+	    DNS_SET_H_ERRNO(&dns_res_state, NO_DATA);
 	    return (status);
 	case DNS_OK:
 	    if (rrlist && dns_rr_filter_maps) {
@@ -1118,129 +1140,16 @@ int     dns_lookup_x(const char *name, unsigned type, unsigned flags,
     return (DNS_NOTFOUND);
 }
 
-/* dns_lookup_rl - DNS lookup interface with types list */
+/* dns_get_h_errno - get the last lookup status */
 
-int     dns_lookup_rl(const char *name, unsigned flags, DNS_RR **rrlist,
-		              VSTRING *fqdn, VSTRING *why, int *rcode,
-		              int lflags,...)
+int     dns_get_h_errno(void)
 {
-    va_list ap;
-    unsigned type, next;
-    int     status = DNS_NOTFOUND;
-    int     hpref_status = INT_MIN;
-    VSTRING *hpref_rtext = 0;
-    int     hpref_rcode;
-    int     hpref_h_errno;
-    DNS_RR *rr;
-
-    /* Save intermediate highest-priority result. */
-#define SAVE_HPREF_STATUS() do { \
-	hpref_status = status; \
-	if (rcode) \
-	    hpref_rcode = *rcode; \
-	if (why && status != DNS_OK) \
-	    vstring_strcpy(hpref_rtext ? hpref_rtext : \
-			   (hpref_rtext = vstring_alloc(VSTRING_LEN(why))), \
-			   vstring_str(why)); \
-	hpref_h_errno = DNS_RES_H_ERRNO(&dns_res_state); \
-    } while (0)
-
-    /* Restore intermediate highest-priority result. */
-#define RESTORE_HPREF_STATUS() do { \
-	status = hpref_status; \
-	if (rcode) \
-	    *rcode = hpref_rcode; \
-	if (why && status != DNS_OK) \
-	    vstring_strcpy(why, vstring_str(hpref_rtext)); \
-	DNS_RES_SET_H_ERRNO(&dns_res_state, hpref_h_errno); \
-    } while (0)
-
-    if (rrlist)
-	*rrlist = 0;
-    va_start(ap, lflags);
-    for (type = va_arg(ap, unsigned); type != 0; type = next) {
-	next = va_arg(ap, unsigned);
-	if (msg_verbose)
-	    msg_info("lookup %s type %s flags %s",
-		     name, dns_strtype(type), dns_str_resflags(flags));
-	status = dns_lookup_x(name, type, flags, rrlist ? &rr : (DNS_RR **) 0,
-			      fqdn, why, rcode, lflags);
-	if (rrlist && rr)
-	    *rrlist = dns_rr_append(*rrlist, rr);
-	if (status == DNS_OK) {
-	    if (lflags & DNS_REQ_FLAG_STOP_OK)
-		break;
-	} else if (status == DNS_INVAL) {
-	    if (lflags & DNS_REQ_FLAG_STOP_INVAL)
-		break;
-	} else if (status == DNS_POLICY) {
-	    if (type == T_MX && (lflags & DNS_REQ_FLAG_STOP_MX_POLICY))
-		break;
-	} else if (status == DNS_NULLMX) {
-	    if (lflags & DNS_REQ_FLAG_STOP_NULLMX)
-		break;
-	}
-	/* XXX Stop after NXDOMAIN error. */
-	if (next == 0)
-	    break;
-	if (status >= hpref_status)
-	    SAVE_HPREF_STATUS();		/* save last info */
-    }
-    va_end(ap);
-    if (status < hpref_status)
-	RESTORE_HPREF_STATUS();			/* else report last info */
-    if (hpref_rtext)
-	vstring_free(hpref_rtext);
-    return (status);
+    return (DNS_GET_H_ERRNO(&dns_res_state));
 }
 
-/* dns_lookup_rv - DNS lookup interface with types vector */
+/* dns_set_h_errno - set the last lookup status */
 
-int     dns_lookup_rv(const char *name, unsigned flags, DNS_RR **rrlist,
-		              VSTRING *fqdn, VSTRING *why, int *rcode,
-		              int lflags, unsigned *types)
+void    dns_set_h_errno(int errval)
 {
-    unsigned type, next;
-    int     status = DNS_NOTFOUND;
-    int     hpref_status = INT_MIN;
-    VSTRING *hpref_rtext = 0;
-    int     hpref_rcode;
-    int     hpref_h_errno;
-    DNS_RR *rr;
-
-    if (rrlist)
-	*rrlist = 0;
-    for (type = *types++; type != 0; type = next) {
-	next = *types++;
-	if (msg_verbose)
-	    msg_info("lookup %s type %s flags %s",
-		     name, dns_strtype(type), dns_str_resflags(flags));
-	status = dns_lookup_x(name, type, flags, rrlist ? &rr : (DNS_RR **) 0,
-			      fqdn, why, rcode, lflags);
-	if (rrlist && rr)
-	    *rrlist = dns_rr_append(*rrlist, rr);
-	if (status == DNS_OK) {
-	    if (lflags & DNS_REQ_FLAG_STOP_OK)
-		break;
-	} else if (status == DNS_INVAL) {
-	    if (lflags & DNS_REQ_FLAG_STOP_INVAL)
-		break;
-	} else if (status == DNS_POLICY) {
-	    if (type == T_MX && (lflags & DNS_REQ_FLAG_STOP_MX_POLICY))
-		break;
-	} else if (status == DNS_NULLMX) {
-	    if (lflags & DNS_REQ_FLAG_STOP_NULLMX)
-		break;
-	}
-	/* XXX Stop after NXDOMAIN error. */
-	if (next == 0)
-	    break;
-	if (status >= hpref_status)
-	    SAVE_HPREF_STATUS();		/* save last info */
-    }
-    if (status < hpref_status)
-	RESTORE_HPREF_STATUS();			/* else report last info */
-    if (hpref_rtext)
-	vstring_free(hpref_rtext);
-    return (status);
+    DNS_SET_H_ERRNO(&dns_res_state, errval);
 }
